#    Copyright 2022 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os
import pytest

from si_tests import logger
from si_tests import settings
from si_tests.managers import bootstrap_manager
from si_tests.managers.kaas_manager import Manager
from si_tests.utils import waiters

LOG = logger.logger


class TestProvisionRegionalClusters():

    @pytest.mark.usefixtures("introspect_standalone_seed_regional_deploy_objects")
    def test_bootstrapv2_em2_regional(self, kaas_manager: Manager, show_step):
        """Deploy Bootstrap V2 Regional EquinixMetalV2 cluster on KaaS mgmt

        Scenario:
            1. Check if a provider is eligible to start
            2. Prepare templates to apply on bootstrap cluster
            3. Copy management kubeconfig
            4. Start bootstrap V2 regional cluster deployment
            5. Check result of deployment
        """

        # step 001: Check if a provider is eligible to start
        show_step(1)
        if not settings.KAAS_EQUINIXMETALV2_ENABLED:
            pytest.fail("Current scenario supports only EquinixMetalV2 provider")

        seed_ip = settings.SEED_STANDALONE_EXTERNAL_IP
        assert waiters.icmp_ping(settings.SEED_STANDALONE_EXTERNAL_IP)

        # step 002: Prepare templates to apply on bootstrap cluster
        show_step(2)
        bootstrap = bootstrap_manager.BootstrapManager(seed_ip=seed_ip)
        bootstrap.step_003_prepare_seed_node_templates()
        bootstrap.step_003_patch_bootstrapv2_release_refs()

        template_path = 'templates.region/equinixmetalv2'
        if os.path.exists(
                os.path.join(bootstrap.get_release_templates_path(remote=bootstrap.remote_seed()),
                             'equinixmetalv2', 'regional')):
            template_path = 'templates/equinixmetalv2/regional'

        # In bootstrap v2 the proxy object is pre-created in the bootstrap,
        # so we need to set that proxy ref in cluster object
        bootstrap.equinixmetalv2_patch_templates(settings.CLUSTER_NAMESPACE,
                                                 settings.REGIONAL_CLUSTER_NAME,
                                                 kaas_manager.si_config.data['network_config'],
                                                 settings.KAAS_MANAGEMENT_CLUSTER_NODES,
                                                 template_path,
                                                 'bootstrap-proxy')

        # step 003: Copy management kubeconfig
        show_step(3)
        bootstrap.copy_user_file(
          os.path.join(settings.SI_BINARIES_DIR, "kubeconfig"),
          settings.SEED_SSH_LOGIN,
          f"/home/{settings.SEED_SSH_LOGIN}/kubeconfig")

        # step 004: Start bootstrap V2 regional cluster deployment
        show_step(4)
        # NOTE: region isn't required in bv2 scenario
        bootstrap.step_004_deploy_kaas_regional(
            region="",
            extra_config=kaas_manager.si_config.data,
            endpoints=None)

        # step 005: Check result of deployment
        show_step(5)
        bootstrap.step_006_wait_for_pods()
        bootstrap.step_006_postdeployment_checks()

        ns = kaas_manager.get_namespace(settings.REGION_NAMESPACE)
        regional_cluster = ns.get_cluster(settings.REGIONAL_CLUSTER_NAME)
        regional_cluster.check.check_machines_status()
        regional_cluster.check.check_cluster_nodes(timeout=900)
        regional_cluster.check.check_helmbundles()
        regional_cluster.check.check_persistent_volumes_mounts()

        regional_cluster.store_k8s_artifacts()
        regional_cluster.store_machines_artifacts(public_ip=False)

        # Check/wait for correct docker service replicas in cluster
        regional_cluster.check.check_actual_expected_docker_services()
        regional_cluster.check.check_k8s_pods()
        regional_cluster.check.check_cluster_readiness()
        regional_cluster.check.check_actual_expected_pods(timeout=1200)
        regional_cluster.provider_resources.save_artifact()
